"""Tests for PEP 649 (Deferred Evaluation Of Annotations Using Descriptors)."""

from __future__ import annotations

from typing import Generic, TypeVar

from attrs import define

from cattrs import Converter

T = TypeVar("T")


@define
class GenericClass(Generic[T]):
    t: T


def test_generics_with_stringified_annotations():
    """Type resolution works with stringified annotations."""
    converter = Converter()
    inst = GenericClass(42)
    dct = converter.unstructure(inst, unstructure_as=GenericClass[int])
    assert dct == {"t": 42}
    assert converter.structure(dct, GenericClass[int])



"""Benchmark attrs containing other attrs classes."""

import pytest
from attrs import define

from cattrs import BaseConverter, Converter, UnstructureStrategy


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter])
@pytest.mark.parametrize(
    "unstructure_strat", [UnstructureStrategy.AS_DICT, UnstructureStrategy.AS_TUPLE]
)
def test_unstructure_attrs_nested(benchmark, converter_cls, unstructure_strat):
    c = converter_cls(unstruct_strat=unstructure_strat)

    @define
    class InnerA:
        a: int
        b: float
        c: str
        d: bytes

    @define
    class InnerB:
        a: int
        b: float
        c: str
        d: bytes

    @define
    class InnerC:
        a: int
        b: float
        c: str
        d: bytes

    @define
    class InnerD:
        a: int
        b: float
        c: str
        d: bytes

    @define
    class InnerE:
        a: int
        b: float
        c: str
        d: bytes

    @define
    class Outer:
        a: InnerA
        b: InnerB
        c: InnerC
        d: InnerD
        e: InnerE

    inst = Outer(
        InnerA(1, 1.0, "one", b"one"),
        InnerB(2, 2.0, "two", b"two"),
        InnerC(3, 3.0, "three", b"three"),
        InnerD(4, 4.0, "four", b"four"),
        InnerE(5, 5.0, "five", b"five"),
    )

    benchmark(c.unstructure, inst)


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter])
@pytest.mark.parametrize(
    "unstructure_strat", [UnstructureStrategy.AS_DICT, UnstructureStrategy.AS_TUPLE]
)
def test_unstruct_attrs_deep_nest(benchmark, converter_cls, unstructure_strat):
    c = converter_cls(unstruct_strat=unstructure_strat)

    @define
    class InnerA:
        a: int
        b: float
        c: str
        d: bytes

    @define
    class InnerB:
        a: InnerA
        b: InnerA
        c: InnerA
        d: InnerA

    @define
    class InnerC:
        a: InnerB
        b: InnerB
        c: InnerB
        d: InnerB

    @define
    class InnerD:
        a: InnerC
        b: InnerC
        c: InnerC
        d: InnerC

    @define
    class InnerE:
        a: InnerD
        b: InnerD
        c: InnerD
        d: InnerD

    @define
    class Outer:
        a: InnerE
        b: InnerE
        c: InnerE
        d: InnerE

    def make_inner_a():
        return InnerA(1, 1.0, "one", b"one")

    def make_inner_b():
        return InnerB(*[make_inner_a() for _ in range(4)])

    def make_inner_c():
        return InnerC(*[make_inner_b() for _ in range(4)])

    def make_inner_d():
        return InnerD(*[make_inner_c() for _ in range(4)])

    def make_inner_e():
        return InnerE(*[make_inner_d() for _ in range(4)])

    inst = Outer(*[make_inner_e() for _ in range(4)])

    benchmark(c.unstructure, inst)



"""Tests for test helpers."""

import pytest
from attrs import define

from .helpers import assert_only_unstructured


def test_assert_only_unstructured_passes_for_primitives():
    """assert_only_unstructured should pass for basic Python data types."""
    # Test primitives
    assert_only_unstructured(42)
    assert_only_unstructured("hello")
    assert_only_unstructured(3.14)
    assert_only_unstructured(True)
    assert_only_unstructured(None)

    # Test collections of primitives
    assert_only_unstructured([1, 2, 3])
    assert_only_unstructured({"key": "value", "number": 42})
    assert_only_unstructured((1, "two", 3.0))
    assert_only_unstructured({1, 2, 3})
    assert_only_unstructured(frozenset([1, 2, 3]))

    # Test nested structures
    assert_only_unstructured(
        {"list": [1, 2, {"nested": "dict"}], "tuple": (True, None), "number": 42}
    )


def test_assert_only_unstructured_fails_for_attrs_classes():
    """assert_only_unstructured should fail for attrs classes."""

    @define
    class SimpleAttrsClass:
        value: int

    instance = SimpleAttrsClass(42)

    # Should raise AssertionError for attrs class instance
    with pytest.raises(AssertionError):
        assert_only_unstructured(instance)

    # Should also fail when attrs instance is nested in collections
    with pytest.raises(AssertionError):
        assert_only_unstructured([instance])

    with pytest.raises(AssertionError):
        assert_only_unstructured({"key": instance})

    with pytest.raises(AssertionError):
        assert_only_unstructured((1, instance, 3))



"""Strategies for typed dicts."""

from datetime import datetime, timezone
from string import ascii_lowercase
from typing import Any, Generic, List, Optional, TypedDict, TypeVar

from attrs import NOTHING
from hypothesis import note
from hypothesis.strategies import (
    DrawFn,
    SearchStrategy,
    booleans,
    composite,
    datetimes,
    integers,
    just,
    lists,
    sets,
    text,
)

from cattrs._compat import Annotated, ExtensionsTypedDict, NotRequired, Required

from .untyped import gen_attr_names

# Type aliases for readability
TypedDictType = type
T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")


def gen_typeddict_attr_names():
    """Typed dicts can have periods in their field names."""
    for counter, n in enumerate(gen_attr_names()):
        if counter % 2 == 0:
            n = f"{n}.suffix"

        yield n


@composite
def int_attributes(
    draw: DrawFn, total: bool = True, not_required: bool = False
) -> tuple[type[int], SearchStrategy, SearchStrategy]:
    if total:
        if not_required and draw(booleans()):
            return (NotRequired[int], integers() | just(NOTHING), text(ascii_lowercase))
        return int, integers(), text(ascii_lowercase)

    if not_required and draw(booleans()):
        return Required[int], integers(), text(ascii_lowercase)

    return int, integers() | just(NOTHING), text(ascii_lowercase)


@composite
def annotated_int_attributes(
    draw: DrawFn, total: bool = True, not_required: bool = False
) -> tuple[int, SearchStrategy, SearchStrategy]:
    """Generate combinations of Annotated types."""
    if total:
        if not_required and draw(booleans()):
            return (
                (
                    NotRequired[Annotated[int, "test"]]
                    if draw(booleans())
                    else Annotated[NotRequired[int], "test"]
                ),
                integers() | just(NOTHING),
                text(ascii_lowercase),
            )
        return Annotated[int, "test"], integers(), text(ascii_lowercase)

    if not_required and draw(booleans()):
        return (
            (
                Required[Annotated[int, "test"]]
                if draw(booleans())
                else Annotated[Required[int], "test"]
            ),
            integers(),
            text(ascii_lowercase),
        )

    return Annotated[int, "test"], integers() | just(NOTHING), text(ascii_lowercase)


@composite
def datetime_attributes(
    draw: DrawFn, total: bool = True, not_required: bool = False
) -> tuple[datetime, SearchStrategy, SearchStrategy]:
    success_strat = datetimes(
        min_value=datetime(1970, 1, 1),
        max_value=datetime(2038, 1, 1),
        timezones=just(timezone.utc),
    ).map(lambda dt: dt.replace(microsecond=0))
    type = datetime
    strat = success_strat if total else success_strat | just(NOTHING)
    if not_required and draw(booleans()):
        if total:
            type = NotRequired[type]
            strat = success_strat | just(NOTHING)
        else:
            type = Required[type]
            strat = success_strat
    return (type, strat, text(ascii_lowercase))


@composite
def list_of_int_attributes(
    draw: DrawFn, total: bool = True, not_required: bool = False
) -> tuple[list[int], SearchStrategy, SearchStrategy]:
    if total:
        if not_required and draw(booleans()):
            return (
                NotRequired[List[int]],
                lists(integers()) | just(NOTHING),
                text(ascii_lowercase).map(lambda v: [v]),
            )

        return (List[int], lists(integers()), text(ascii_lowercase).map(lambda v: [v]))

    if not_required and draw(booleans()):
        return (
            Required[List[int]],
            lists(integers()),
            text(ascii_lowercase).map(lambda v: [v]),
        )

    return (
        List[int],
        lists(integers()) | just(NOTHING),
        text(ascii_lowercase).map(lambda v: [v]),
    )


@composite
def simple_typeddicts(
    draw: DrawFn,
    total: Optional[bool] = None,
    not_required: bool = False,
    min_attrs: int = 0,
    typeddict_cls: Optional[Any] = None,
) -> tuple[TypedDictType, dict]:
    """Generate simple typed dicts.

    :param total: Generate the given totality dicts (default = random)
    """
    if total is None:
        total = draw(booleans())

    attrs = draw(
        lists(
            int_attributes(total, not_required)
            | annotated_int_attributes(total, not_required)
            | list_of_int_attributes(total, not_required)
            | datetime_attributes(total, not_required),
            min_size=min_attrs,
        )
    )

    attrs_dict = {n: attr[0] for n, attr in zip(gen_typeddict_attr_names(), attrs)}
    success_payload = {}
    for n, a in zip(attrs_dict, attrs):
        v = draw(a[1])
        if v is not NOTHING:
            success_payload[n] = v

    cls = (
        (TypedDict if draw(booleans()) else ExtensionsTypedDict)
        if typeddict_cls is None
        else typeddict_cls
    )("HypTypedDict", attrs_dict, total=total)

    note(
        "\n".join(
            [
                f"class HypTypedDict(TypedDict{'' if total else ', total=False'}):",
                *[f"    {n}: {a}" for n, a in attrs_dict.items()],
            ]
        )
    )

    if draw(booleans()):

        class InheritedTypedDict(cls):
            inherited: int

        cls = InheritedTypedDict
        success_payload["inherited"] = draw(integers())

    return (cls, success_payload)


@composite
def simple_typeddicts_with_extra_keys(
    draw: DrawFn, total: Optional[bool] = None, typeddict_cls: Optional[Any] = None
) -> tuple[TypedDictType, dict, set[str]]:
    """Generate TypedDicts, with the instances having extra keys."""
    cls, success = draw(simple_typeddicts(total, typeddict_cls=typeddict_cls))

    # The normal attributes are 2 characters or fewer.
    extra_keys = draw(sets(text(ascii_lowercase, min_size=3, max_size=3)))
    success.update(dict.fromkeys(extra_keys, 1))

    return cls, success, extra_keys


@composite
def generic_typeddicts(draw: DrawFn, total: bool = True) -> tuple[TypedDictType, dict]:
    """Generate generic typed dicts.

    :param total: Generate the given totality dicts
    """
    attrs = draw(
        lists(
            int_attributes(total)
            | list_of_int_attributes(total)
            | datetime_attributes(total),
            min_size=1,
        )
    )

    attrs_dict = {n: attr[0] for n, attr in zip(gen_attr_names(), attrs)}
    success_payload = {}
    for n, a in zip(attrs_dict, attrs):
        v = draw(a[1])
        if v is not NOTHING:
            success_payload[n] = v

    # We choose up to 3 attributes and make them generic.
    generic_attrs = draw(
        lists(integers(0, len(attrs) - 1), min_size=1, max_size=3, unique=True)
    )
    generics = []
    actual_types = []
    for ix, (attr_name, attr_type) in enumerate(list(attrs_dict.items())):
        if ix in generic_attrs:
            typevar = TypeVar(f"T{ix+1}")
            generics.append(typevar)
            if total and draw(booleans()):
                # We might decide to make these NotRequired
                typevar = NotRequired[typevar]
            actual_types.append(attr_type)
            attrs_dict[attr_name] = typevar

    cls = make_typeddict(
        "HypTypedDict", attrs_dict, total=total, bases=[Generic[tuple(generics)]]
    )

    if draw(booleans()):

        class InheritedTypedDict(cls[tuple(actual_types)]):
            inherited: int

        cls = InheritedTypedDict
        success_payload["inherited"] = draw(integers())
    else:
        cls = cls[tuple(actual_types)]

    return (cls, success_payload)


def make_typeddict(
    cls_name: str, attrs: dict[str, type], total: bool = True, bases: list = []
) -> TypedDictType:
    globs = {"TypedDict": TypedDict}
    lines = []

    bases_snippet = ", ".join(f"_base{ix}" for ix in range(len(bases)))
    for ix, base in enumerate(bases):
        globs[f"_base{ix}"] = base
    bases_snippet = f", {bases_snippet}"

    lines.append(f"class {cls_name}(TypedDict{bases_snippet}, total={total}):")
    for n, t in attrs.items():
        # Strip the initial underscore if present, to prevent mangling.
        trimmed = n[1:] if n.startswith("_") else n
        globs[f"_{trimmed}_type"] = t
        lines.append(f"  {n}: _{trimmed}_type")

    script = "\n".join(lines)

    note_lines = script
    for n, t in globs.items():
        if n == "TypedDict":
            continue
        note_lines = note_lines.replace(n, repr(t))
    note(note_lines)

    eval(compile(script, "name", "exec"), globs)

    return globs[cls_name]



from .untyped import gen_attr_names


def test_gen_attr_names():
    """We can generate a lot of attribute names."""
    assert len(list(gen_attr_names())) == 697

    # No duplicates!
    assert len(list(gen_attr_names())) == len(set(gen_attr_names()))



"""Tests for defaultdicts."""

from collections import defaultdict
from typing import DefaultDict

from cattrs import Converter
from cattrs.cols import defaultdict_structure_factory


def test_typing_defaultdicts(genconverter: Converter):
    """`typing.DefaultDict` works."""
    res = genconverter.structure({"a": 1}, DefaultDict[str, int])

    assert isinstance(res, defaultdict)
    assert res["a"] == 1
    assert res["b"] == 0

    genconverter.register_unstructure_hook(int, str)

    assert genconverter.unstructure(res) == {"a": "1", "b": "0"}


def test_collection_defaultdicts(genconverter: Converter):
    """`collections.defaultdict` works."""
    res = genconverter.structure({"a": 1}, defaultdict[str, int])

    assert isinstance(res, defaultdict)
    assert res["a"] == 1
    assert res["b"] == 0

    genconverter.register_unstructure_hook(int, str)

    assert genconverter.unstructure(res) == {"a": "1", "b": "0"}


def test_factory(genconverter: Converter):
    """Explicit factories work."""
    genconverter.register_structure_hook_func(
        lambda t: t == defaultdict[str, int],
        defaultdict_structure_factory(defaultdict[str, int], genconverter, lambda: 2),
    )
    res = genconverter.structure({"a": 1}, defaultdict[str, int])

    assert isinstance(res, defaultdict)
    assert res["a"] == 1
    assert res["b"] == 2



from dataclasses import dataclass
from typing import Annotated, NamedTuple, TypedDict

from attrs import define

from cattrs import Converter
from cattrs._compat import NotRequired
from cattrs.cols import (
    is_namedtuple,
    namedtuple_dict_structure_factory,
    namedtuple_dict_unstructure_factory,
)
from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn, override
from cattrs.gen.typeddicts import make_dict_structure_fn as make_td_structure_fn
from cattrs.gen.typeddicts import make_dict_unstructure_fn as make_td_unstructure_fn


def test_annotated_override_attrs(genconverter: Converter):
    """Annotated overrides work for attrs classes."""

    @define
    class A:
        a: Annotated[int, override(rename="b")]
        c: Annotated[int, override(omit=True)] = 1
        d: Annotated[int, override(rename="e")] = 2

    instance = A(1)
    # 'a' is renamed to 'b', 'c' is omitted. 'd' is default so present as 'e'
    assert genconverter.unstructure(instance) == {"b": 1, "e": 2}

    assert genconverter.structure({"b": 1, "e": 2}, A) == A(1)


def test_annotated_override_dataclasses(genconverter: Converter):
    """Annotated overrides work for dataclasses."""

    @dataclass
    class A:
        a: Annotated[int, override(rename="b")]
        c: Annotated[int, override(omit=True)] = 1

    instance = A(1)
    assert genconverter.unstructure(instance) == {"b": 1}

    assert genconverter.structure({"b": 1}, A) == A(1)


def test_annotated_override_typeddict(genconverter: Converter):
    """Annotated overrides work for TypedDicts."""

    class TD(TypedDict):
        a: Annotated[int, override(rename="b")]
        c: Annotated[int, override(omit=True)]

    instance: TD = {"a": 1, "c": 2}

    assert genconverter.unstructure(instance, TD) == {"b": 1}

    # Let's simplify and just test rename for now to avoid required field issues with omit.
    class TD2(TypedDict):
        a: Annotated[int, override(rename="b")]

    inst2: TD2 = {"a": 1}
    assert genconverter.unstructure(inst2, TD2) == {"b": 1}
    assert genconverter.structure({"b": 1}, TD2) == {"a": 1}


def test_annotated_override_namedtuple(genconverter: Converter):
    """Annotated overrides work for NamedTuples using dict factories."""

    # We need to register the dict factories for NamedTuples
    genconverter.register_unstructure_hook_factory(
        is_namedtuple, namedtuple_dict_unstructure_factory
    )
    genconverter.register_structure_hook_factory(
        is_namedtuple, namedtuple_dict_structure_factory
    )

    class NT(NamedTuple):
        a: Annotated[int, override(rename="b")]
        c: Annotated[int, override(omit=True)] = 1

    instance = NT(1)
    assert genconverter.unstructure(instance) == {"b": 1}
    assert genconverter.structure({"b": 1}, NT) == NT(1)


def test_annotated_override_precedence(genconverter: Converter):
    """Test that explicit kwargs override Annotated metadata."""

    @define
    class A:
        a: Annotated[int, override(rename="b")]

    # Override the rename back to 'a' explicitly
    unstructure_fn = make_dict_unstructure_fn(A, genconverter, a=override(rename="a"))
    genconverter.register_unstructure_hook(A, unstructure_fn)

    assert genconverter.unstructure(A(1)) == {"a": 1}

    # # Structure override
    structure_fn = make_dict_structure_fn(A, genconverter, a=override(rename="a"))
    genconverter.register_structure_hook(A, structure_fn)

    assert genconverter.structure({"a": 1}, A) == A(1)


def test_annotated_override_hooks(genconverter: Converter):
    """struct_hook and unstruct_hook work in Annotated."""

    def double_hook(v):
        return v * 2

    def half_hook(v, _):
        return v // 2

    @define
    class A:
        a: Annotated[int, override(unstruct_hook=double_hook, struct_hook=half_hook)]

    assert genconverter.unstructure(A(10)) == {"a": 20}
    assert genconverter.structure({"a": 20}, A) == A(10)


def test_annotated_override_omit_if_default(genconverter: Converter):
    """omit_if_default works in Annotated."""

    @define
    class A:
        a: Annotated[int, override(omit_if_default=True)] = 0
        b: int = 1

    # a matches default, should be omitted. b matches default but no override, should stay (default behavior is to keep)
    assert genconverter.unstructure(A()) == {"b": 1}
    assert genconverter.unstructure(A(a=1)) == {"a": 1, "b": 1}


def test_overrides_attribute_populated(genconverter: Converter):
    """The .overrides attribute is correctly populated."""

    @dataclass
    class A:
        a: Annotated[int, override(rename="b")]
        c: Annotated[int, override(omit=True)] = 1

    # Test dataclasses (make_dict_unstructure_fn)
    unstruct_hook = make_dict_unstructure_fn(A, genconverter)
    assert unstruct_hook.overrides == {
        "a": override(rename="b"),
        "c": override(omit=True),
    }

    struct_hook = make_dict_structure_fn(A, genconverter)
    assert struct_hook.overrides == {
        "a": override(rename="b"),
        "c": override(omit=True),
    }

    class TD(TypedDict):
        a: Annotated[int, override(rename="b")]
        c: NotRequired[Annotated[int, override(rename="d")]]

    td_unstruct_hook = make_td_unstructure_fn(TD, genconverter)
    assert td_unstruct_hook.overrides == {
        "a": override(rename="b"),
        "c": override(rename="d"),
    }

    td_struct_hook = make_td_structure_fn(TD, genconverter)
    assert td_struct_hook.overrides == {
        "a": override(rename="b"),
        "c": override(rename="d"),
    }

    # Test Precedence (explicit should win and be in overrides)
    @dataclass
    class B:
        a: Annotated[int, override(rename="b")]

    hook_explicit = make_dict_unstructure_fn(B, genconverter, a=override(rename="c"))
    assert hook_explicit.overrides["a"].rename == "c"



from enum import Enum
from typing import Literal

from cattrs import BaseConverter
from cattrs.fns import identity


class AnEnum(Enum):
    TEST = "test"


def test_unstructure_literal(converter: BaseConverter):
    """Literals without enums are passed through by default."""
    assert converter.get_unstructure_hook(1, Literal[1]) == identity


def test_unstructure_literal_with_enum(converter: BaseConverter):
    """Literals with enums are properly unstructured."""
    assert converter.unstructure(AnEnum.TEST, Literal[AnEnum.TEST]) == "test"



import pickle
from pathlib import Path
from typing import Any

import pytest

from cattrs._compat import ExceptionGroup
from cattrs.errors import (
    BaseValidationError,
    ClassValidationError,
    ForbiddenExtraKeysError,
    IterableValidationError,
    StructureHandlerNotFoundError,
)


@pytest.mark.parametrize(
    "err_cls, err_args",
    [
        (StructureHandlerNotFoundError, ("Structure Message", int)),
        (ForbiddenExtraKeysError, ("Forbidden Message", int, {"foo", "bar"})),
        (ForbiddenExtraKeysError, ("", str, {"foo", "bar"})),
        (ForbiddenExtraKeysError, (None, list, {"foo", "bar"})),
        (
            BaseValidationError,
            ("BaseValidation Message", [ValueError("Test BaseValidation")], int),
        ),
        (
            IterableValidationError,
            ("IterableValidation Msg", [ValueError("Test IterableValidation")], int),
        ),
        (
            ClassValidationError,
            ("ClassValidation Message", [ValueError("Test ClassValidation")], int),
        ),
    ],
)
def test_errors_pickling(
    err_cls: type[Exception], err_args: tuple[Any, ...], tmp_path: Path
) -> None:
    """Test if a round of pickling and unpickling works for errors."""
    before = err_cls(*err_args)

    assert before.args == err_args
    after = pickle.loads(pickle.dumps(before))  # noqa: S301

    assert isinstance(after, err_cls)

    assert str(after) == str(before)

    if issubclass(err_cls, ExceptionGroup):
        assert after.message == before.message
        assert after.args[0] == before.args[0]

        # We need to do the exceptions within the group (i.e. args[1])
        # separately, as on unpickling new objects are created and hence
        # they will never be equal to the original ones.
        for after_exc, before_exc in zip(after.exceptions, before.exceptions):
            assert str(after_exc) == str(before_exc)

        # The problem with args[1] might be also for other parameters, but
        # we ignore this here and if needed then we need a separate test
        assert after.args[2:] == before.args[2:]

    else:
        assert after.args == err_args
        assert after.args == before.args

    assert after.__cause__ == before.__cause__
    assert after.__context__ == before.__context__
    assert after.__traceback__ == before.__traceback__



from __future__ import annotations

from attrs import define

from cattrs import BaseConverter, Converter
from cattrs.strategies import configure_tagged_union

from .test_tagged_unions import A, B


def test_type_alias(converter: BaseConverter):
    """Type aliases to unions also work."""
    type AOrB = A | B

    configure_tagged_union(AOrB, converter)

    assert converter.unstructure(A(1), AOrB) == {"_type": "A", "a": 1}
    assert converter.unstructure(B("1"), AOrB) == {"_type": "B", "a": "1"}

    assert converter.structure({"_type": "A", "a": 1}, AOrB) == A(1)
    assert converter.structure({"_type": "B", "a": 1}, AOrB) == B("1")


@define
class Lit:
    value: float


@define
class Add:
    left: Expr
    right: Expr


type Expr = Add | Lit


def test_recursive_type_alias(genconverter: Converter):
    """Recursive type aliases to unions also work.

    Only tests on the GenConverter since the BaseConverter doesn't support
    stringified annotations.
    """

    configure_tagged_union(Expr, genconverter)

    val = Add(Lit(1.0), Lit(2.0))
    expected = {
        "_type": "Add",
        "left": {"_type": "Lit", "value": 1.0},
        "right": {"_type": "Lit", "value": 2.0},
    }

    assert genconverter.unstructure(val, Expr) == expected
    assert genconverter.structure(expected, Expr) == val



"""Tests for enums."""

from enum import Enum

from hypothesis import given
from hypothesis.strategies import data, sampled_from
from pytest import raises

from cattrs import BaseConverter
from cattrs._compat import Literal

from .untyped import enums_of_primitives


@given(data(), enums_of_primitives())
def test_structuring_enums(data, enum):
    """Test structuring enums by their values."""
    converter = BaseConverter()
    val = data.draw(sampled_from(list(enum)))

    assert converter.structure(val.value, enum) == val


@given(enums_of_primitives())
def test_enum_failure(enum):
    """Structuring literals with enums fails properly."""
    converter = BaseConverter()
    type = Literal[next(iter(enum))]

    with raises(Exception) as exc_info:
        converter.structure("", type)

    assert exc_info.value.args[0] == f" not in literal {type!r}"


class SimpleEnum(Enum):
    A = 0
    B = 1
    C = 2


class SimpleEnumWithTypeHint(Enum):
    _value_: str
    D = "D"
    E = "E"
    F = "F"


class ComplexEnum(Enum):
    _value_: tuple[SimpleEnum, SimpleEnumWithTypeHint]
    AD = (SimpleEnum.A, SimpleEnumWithTypeHint.D)
    AE = (SimpleEnum.A, SimpleEnumWithTypeHint.E)
    BE = (SimpleEnum.B, SimpleEnumWithTypeHint.E)
    BF = (SimpleEnum.B, SimpleEnumWithTypeHint.F)
    CE = (SimpleEnum.C, SimpleEnumWithTypeHint.E)


def test_unstructure_complex_enum() -> None:
    converter = BaseConverter()
    assert converter.unstructure(SimpleEnum.A) == 0
    assert converter.unstructure(SimpleEnumWithTypeHint.F) == "F"
    assert converter.unstructure(ComplexEnum.AE) == (0, "E")


def test_structure_complex_enum() -> None:
    converter = BaseConverter()
    assert converter.structure(0, SimpleEnum) == SimpleEnum.A
    assert converter.structure("E", SimpleEnumWithTypeHint) == SimpleEnumWithTypeHint.E
    assert converter.structure((0, "D"), ComplexEnum) == ComplexEnum.AD



import dataclasses
from typing import List

import attr
import pytest

from cattrs import BaseConverter

from ._compat import is_py310_plus


@dataclasses.dataclass
class Foo:
    bar: str


@attr.define
class Container:
    foos: List[Foo]


def test_dataclasses_in_attrs(converter: BaseConverter):
    struct = Container([Foo("bar")])

    unstruct = {"foos": [{"bar": "bar"}]}

    assert converter.unstructure(struct) == unstruct
    assert converter.structure(unstruct, Container) == struct


def test_dataclasses_in_container(converter: BaseConverter):
    struct = [Foo("bar"), Foo("bat")]

    unstruct = [{"bar": "bar"}, {"bar": "bat"}]

    assert converter.unstructure(struct) == unstruct
    assert converter.structure(unstruct, List[Foo]) == struct


def test_dataclasses(converter: BaseConverter):
    struct = Foo("bar")

    unstruct = {"bar": "bar"}

    assert converter.unstructure(struct) == unstruct
    assert converter.structure(unstruct, Foo) == struct


@pytest.mark.skipif(not is_py310_plus, reason="kwonly fields are Python 3.10+")
def test_kw_only_propagation(converter: BaseConverter):
    """KW-only args work.

    Reproducer from https://github.com/python-attrs/cattrs/issues/637.
    """

    @dataclasses.dataclass
    class PartialKeywords:
        a1: str = "Default"
        a2: str = dataclasses.field(kw_only=True)

    assert converter.structure({"a2": "Value"}, PartialKeywords) == PartialKeywords(
        a1="Default", a2="Value"
    )



from collections import deque
from collections.abc import Sequence
from typing import Deque, Dict, Generic, List, Optional, TypedDict, TypeVar, Union

import pytest
from attrs import asdict, define

from cattrs import BaseConverter, Converter
from cattrs._compat import Protocol
from cattrs._generics import deep_copy_with
from cattrs.errors import StructureHandlerNotFoundError
from cattrs.gen._generics import generate_mapping

from ._compat import Dict_origin, List_origin, is_py310_plus, is_py311_plus

T = TypeVar("T")
T2 = TypeVar("T2")
T3 = TypeVar("T3", bound=int)


def test_deep_copy():
    """Test the deep copying of generic parameters."""
    mapping = {T.__name__: int}
    assert deep_copy_with(Optional[T], mapping) == Optional[int]
    assert (
        deep_copy_with(List_origin[Optional[T]], mapping) == List_origin[Optional[int]]
    )
    mapping = {T.__name__: int, T2.__name__: str}
    assert (
        deep_copy_with(Dict_origin[T2, List_origin[Optional[T]]], mapping)
        == Dict_origin[str, List_origin[Optional[int]]]
    )


@define
class TClass(Generic[T, T2, T3]):
    a: T
    b: T2
    c: T3 = 0


@define
class GenericCols(Generic[T]):
    a: T
    b: List[T]
    c: Dict[str, T]


@pytest.mark.parametrize(
    ("t", "t2", "t3", "result"),
    (
        (int, str, int, TClass(1, "a")),
        (str, str, int, TClass("1", "a")),
        (List[int], str, int, TClass([1, 2, 3], "a")),
    ),
)
def test_able_to_structure_generics(converter: BaseConverter, t, t2, t3, result):
    res = converter.structure(asdict(result), TClass[t, t2, t3])

    assert res == result


@pytest.mark.parametrize(
    ("t", "result"),
    ((int, GenericCols(1, [2], {"3": 3})), (str, GenericCols("1", ["2"], {"3": "3"}))),
)
@pytest.mark.parametrize("detailed_validation", [True, False])
def test_structure_generics_with_cols(t, result, detailed_validation):
    raw = asdict(result)
    res = Converter(detailed_validation=detailed_validation).structure(
        raw, GenericCols[t]
    )

    assert res == result


@pytest.mark.parametrize(
    ("t", "result"), ((int, (1, [2], {"3": 3})), (str, ("1", ["2"], {"3": "3"})))
)
def test_39_structure_generics_with_cols(t, result, genconverter: Converter):
    @define
    class GenericCols(Generic[T]):
        a: T
        b: list[T]
        c: dict[str, T]

    expected = GenericCols(*result)

    res = genconverter.structure(asdict(expected), GenericCols[t])

    assert res == expected


@pytest.mark.parametrize(("t", "result"), ((int, (1, [1, 2, 3])), (int, (1, None))))
def test_structure_nested_generics_with_cols(t, result, genconverter: Converter):
    @define
    class GenericCols(Generic[T]):
        a: T
        b: Optional[List[T]]

    expected = GenericCols(*result)

    res = genconverter.structure(asdict(expected), GenericCols[t])

    assert res == expected


@pytest.mark.parametrize(
    ("t", "t2", "t3", "result"),
    (
        (TClass[int, int, int], str, int, TClass(TClass(1, 2), "a")),
        (List[TClass[int, int, int]], str, int, TClass([TClass(1, 2)], "a")),
        (
            Sequence[TClass[str, str, str]],
            str,
            str,
            TClass((TClass("a", "b", "c"),), "b", "c"),
        ),
    ),
)
def test_structure_nested_generics(converter: BaseConverter, t, t2, t3, result):
    """Structuring nested generics works."""
    res = converter.structure(asdict(result), TClass[t, t2, t3])

    assert res == result


def test_able_to_structure_deeply_nested_generics_gen(converter):
    cl = TClass[TClass[TClass[int, int, int], int, int], int, int]
    result = TClass(TClass(TClass(1, 2), 3), 4)

    res = converter.structure(asdict(result), cl)

    assert res == result


def test_structure_unions_of_generics(converter):
    @define
    class TClass2(Generic[T]):
        c: T

    data = TClass2(c="string")
    res = converter.structure(asdict(data), Union[TClass[int, int, int], TClass2[str]])
    assert res == data


def test_structure_list_of_generic_unions(converter):
    @define
    class TClass2(Generic[T]):
        c: T

    data = [TClass2(c="string"), TClass(1, 2)]
    res = converter.structure(
        [asdict(x) for x in data], List[Union[TClass[int, int, int], TClass2[str]]]
    )
    assert res == data


def test_structure_deque_of_generic_unions(converter):
    @define
    class TClass2(Generic[T]):
        c: T

    data = deque((TClass2(c="string"), TClass(1, 2)))
    res = converter.structure(
        [asdict(x) for x in data], Deque[Union[TClass[int, int, int], TClass2[str]]]
    )
    assert res == data


def test_raises_if_no_generic_params_supplied(
    converter: Union[Converter, BaseConverter],
):
    data = TClass(1, "a")

    with pytest.raises(
        StructureHandlerNotFoundError,
        match="Unsupported type: ~T. Register a structure hook for it.|Missing type for generic argument T, specify it when structuring.",
    ) as exc:
        converter.structure(asdict(data), TClass)

    assert exc.value.type_ is T


def test_unstructure_generic_attrs(genconverter):
    @define
    class Inner(Generic[T]):
        a: T

    inner = Inner(Inner(1))
    assert genconverter.unstructure(inner) == {"a": {"a": 1}}

    @define
    class Outer:
        inner: Inner[int]

    initial = Outer(Inner(1))
    raw = genconverter.unstructure(initial)

    assert raw == {"inner": {"a": 1}}

    new = genconverter.structure(raw, Outer)
    assert initial == new

    @define
    class OuterStr:
        inner: Inner[str]

    assert genconverter.structure(raw, OuterStr) == OuterStr(Inner("1"))


def test_unstructure_generic_inheritance(genconverter):
    """Classes inheriting from generic classes work."""
    genconverter.register_unstructure_hook(int, lambda v: v + 1)
    genconverter.register_unstructure_hook(str, lambda v: str(int(v) + 1))

    @define
    class Parent(Generic[T]):
        a: T

    @define
    class Child(Parent, Generic[T]):
        b: str

    instance = Child(1, "2")
    assert genconverter.unstructure(instance, Child[int]) == {"a": 2, "b": "3"}

    @define
    class ExplicitChild(Parent[int]):
        b: str

    instance = ExplicitChild(1, "2")
    assert genconverter.unstructure(instance, ExplicitChild) == {"a": 2, "b": "3"}


def test_unstructure_optional(genconverter):
    """Generics with optional fields work."""

    @define
    class C(Generic[T]):
        a: Union[T, None]

    assert genconverter.unstructure(C(C(1))) == {"a": {"a": 1}}


def test_unstructure_deeply_nested_generics(genconverter):
    @define
    class Inner:
        a: int

    @define
    class Outer(Generic[T]):
        inner: T

    initial = Outer[Inner](Inner(1))
    raw = genconverter.unstructure(initial, Outer[Inner])
    assert raw == {"inner": {"a": 1}}

    raw = genconverter.unstructure(initial)
    assert raw == {"inner": {"a": 1}}


def test_unstructure_deeply_nested_generics_list(genconverter):
    @define
    class Inner:
        a: int

    @define
    class Outer(Generic[T]):
        inner: List[T]

    initial = Outer[Inner]([Inner(1)])
    raw = genconverter.unstructure(initial, Outer[Inner])
    assert raw == {"inner": [{"a": 1}]}

    raw = genconverter.unstructure(initial)
    assert raw == {"inner": [{"a": 1}]}


def test_unstructure_protocol(genconverter):
    class Proto(Protocol):
        a: int

    @define
    class Inner:
        a: int

    @define
    class Outer:
        inner: Proto

    initial = Outer(Inner(1))
    raw = genconverter.unstructure(initial, Outer)
    assert raw == {"inner": {"a": 1}}

    raw = genconverter.unstructure(initial)
    assert raw == {"inner": {"a": 1}}


@pytest.mark.skipif(not is_py310_plus, reason="3.10+ union syntax")
def test_roundtrip_generic_with_union() -> None:
    """Generators should handle classes with unions in their names."""
    c = Converter()

    @define
    class A:
        a: int

    @define
    class B:
        b: int

    @define
    class Outer(Generic[T]):
        member: T

    raw = c.unstructure(Outer(A(1)), unstructure_as=Outer[A | B])
    assert c.structure(raw, Outer[A | B]) == Outer(A(1))


@pytest.mark.skipif(not is_py311_plus, reason="3.11+ only")
def test_generate_typeddict_mapping() -> None:

    T = TypeVar("T")
    U = TypeVar("U")

    class A(TypedDict):
        pass

    assert generate_mapping(A, {}) == {}

    class A(TypedDict, Generic[T]):
        a: T

    assert generate_mapping(A[int], {}) == {T.__name__: int}

    class B(A[int]):
        pass

    assert generate_mapping(B, {}) == {T.__name__: int}

    class C(Generic[T, U]):
        a: T
        c: U

    assert generate_mapping(C[int, U], {}) == {T.__name__: int}


def test_nongeneric_protocols(converter):
    """Non-generic protocols work."""

    class NongenericProtocol(Protocol): ...

    @define
    class Entity(NongenericProtocol): ...

    assert generate_mapping(Entity) == {}

    class GenericProtocol(Protocol[T]): ...

    @define
    class GenericEntity(GenericProtocol[int]):
        a: int

    assert generate_mapping(GenericEntity) == {"T": int}

    assert converter.structure({"a": 1}, GenericEntity) == GenericEntity(1)



import sys

is_py310 = sys.version_info[:2] == (3, 10)
is_py310_plus = sys.version_info >= (3, 10)
is_py311_plus = sys.version_info >= (3, 11)
is_py312_plus = sys.version_info >= (3, 12)


List_origin = list
Dict_origin = dict



"""Tests un/structure of nested generic classes (stringified only)"""

from __future__ import annotations

from typing import Generic, TypeVar

from attrs import define

T = TypeVar("T")


def test_structure_nested_roundtrip(genconverter):
    @define(auto_attribs=True)
    class Inner:
        value: int

    @define(auto_attribs=True)
    class Container(Generic[T]):
        data: T

    raw = {"data": {"value": 42}}
    structured = genconverter.structure(raw, Container[Inner])
    assert genconverter.unstructure(structured, Container[Inner]) == raw



# ruff: noqa: PLC0415
from collections.abc import Callable, Set
from datetime import date, datetime, timezone
from enum import Enum, IntEnum, unique
from json import dumps as json_dumps
from json import loads as json_loads
from platform import python_implementation
from typing import Any, Dict, Final, List, Literal, NamedTuple, NewType, Union

import pytest
from attrs import define, fields
from bson import CodecOptions, ObjectId
from hypothesis import given, settings
from hypothesis.strategies import (
    DrawFn,
    binary,
    booleans,
    builds,
    characters,
    composite,
    dates,
    datetimes,
    dictionaries,
    floats,
    frozensets,
    integers,
    just,
    lists,
    one_of,
    sampled_from,
    sets,
    text,
)

from cattrs import Converter
from cattrs._compat import (
    Counter,
    FrozenSet,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
    TupleSubscriptable,
)
from cattrs.fns import identity
from cattrs.preconf.bson import make_converter as bson_make_converter
from cattrs.preconf.cbor2 import make_converter as cbor2_make_converter
from cattrs.preconf.json import make_converter as json_make_converter
from cattrs.preconf.msgpack import make_converter as msgpack_make_converter
from cattrs.preconf.pyyaml import make_converter as pyyaml_make_converter
from cattrs.preconf.tomlkit import make_converter as tomlkit_make_converter
from cattrs.preconf.tomllib import make_converter as tomllib_make_converter
from cattrs.preconf.ujson import make_converter as ujson_make_converter

NO_MSGSPEC: Final = python_implementation() == "PyPy"
NO_ORJSON: Final = python_implementation() == "PyPy"


@define
class A:
    a: int


@define
class B:
    b: str


class C(NamedTuple):
    c: float


@define
class Everything:
    @unique
    class AnIntEnum(IntEnum):
        A = 1

    @unique
    class AStringEnum(str, Enum):
        A = "a"

    class ABareEnum(Enum):
        B = "b"

    string: str
    bytes: bytes
    an_int: int
    a_float: float
    a_dict: Dict[str, int]
    a_bare_dict: dict
    a_list: List[int]
    a_homogenous_tuple: TupleSubscriptable[int, ...]
    a_hetero_tuple: TupleSubscriptable[str, int, float]
    a_counter: Counter[str]
    a_mapping: Mapping[int, float]
    a_mutable_mapping: MutableMapping[float, str]
    a_sequence: Sequence[float]
    a_mutable_sequence: MutableSequence[str]
    a_set: Set[float]
    a_mutable_set: MutableSet[int]
    a_frozenset: FrozenSet[str]
    an_int_enum: AnIntEnum
    a_str_enum: AStringEnum
    a_bare_enum: ABareEnum
    a_datetime: datetime
    a_date: date
    a_string_enum_dict: Dict[AStringEnum, int]
    a_bytes_dict: Dict[bytes, bytes]
    native_union: Union[int, float, str]
    native_union_with_spillover: Union[int, str, Set[str]]
    native_union_with_union_spillover: Union[int, str, A, B]
    a_namedtuple: C
    a_literal: Literal[1, AStringEnum.A]
    a_literal_with_bare: Literal[1, ABareEnum.B]


@composite
def everythings(
    draw: DrawFn,
    min_int=None,
    max_int=None,
    allow_inf=True,
    allow_null_bytes_in_keys=True,
    allow_control_characters_in_values=True,
    min_key_length=0,
    allow_datetime_microseconds=True,
    key_blacklist_characters=[],
):
    """
    Args:
        min_key_length: The minimum key length for keys in dicts.
    """
    key_text = text(
        characters(
            blacklist_categories=("Cs",) if allow_null_bytes_in_keys else ("Cs", "Cc"),
            blacklist_characters=key_blacklist_characters,
        ),
        min_size=min_key_length,
    )
    strings = text(
        characters(
            blacklist_categories=(
                ("Cs",) if allow_control_characters_in_values else ("Cs", "Cc")
            )
        )
    )
    dts = datetimes(
        min_value=datetime(1970, 1, 1),
        max_value=datetime(2038, 1, 1),
        timezones=just(timezone.utc),
    )
    if not allow_datetime_microseconds:
        dts = dts.map(
            lambda d: datetime(
                d.year, d.month, d.day, d.hour, d.minute, d.second, tzinfo=d.tzinfo
            )
        )
    fs = floats(allow_nan=False, allow_infinity=allow_inf)
    ints = integers(min_value=min_int, max_value=max_int)

    return Everything(
        string=draw(strings),
        bytes=draw(binary()),
        an_int=draw(ints),
        a_float=draw(fs),
        a_dict=draw(dictionaries(key_text, ints)),
        a_bare_dict=draw(dictionaries(key_text, strings)),
        a_list=draw(lists(ints)),
        a_homogenous_tuple=draw(lists(ints).map(tuple)),
        a_hetero_tuple=(draw(strings), draw(ints), draw(fs)),
        a_counter=Counter(draw(dictionaries(key_text, ints))),
        a_mapping=draw(dictionaries(ints, fs)),
        a_mutable_mapping=draw(dictionaries(fs, strings)),
        a_sequence=draw(lists(fs).map(tuple)),
        a_mutable_sequence=draw(lists(strings)),
        a_set=draw(sets(fs)),
        a_mutable_set=draw(sets(ints)),
        a_frozenset=draw(frozensets(strings)),
        an_int_enum=Everything.AnIntEnum.A,
        a_str_enum=Everything.AStringEnum.A,
        a_bare_enum=Everything.ABareEnum.B,
        a_datetime=draw(dts),
        a_date=draw(dates(min_value=date(1970, 1, 1), max_value=date(2038, 1, 1))),
        a_string_enum_dict=draw(dictionaries(just(Everything.AStringEnum.A), ints)),
        a_bytes_dict=draw(dictionaries(binary(min_size=min_key_length), binary())),
        native_union=draw(one_of(ints, fs, strings)),
        native_union_with_spillover=draw(one_of(ints, strings, sets(strings))),
        native_union_with_union_spillover=draw(
            one_of(ints, strings, ints.map(A), strings.map(B))
        ),
        a_namedtuple=draw(fs.map(C)),
        a_literal=draw(one_of(just(1), just(Everything.AStringEnum.A))),
        a_literal_with_bare=draw(one_of(just(1), just(Everything.ABareEnum.B))),
    )


NewStr = NewType("NewStr", str)
NewInt = NewType("NewInt", int)
NewBool = NewType("NewBool", bool)


@composite
def native_unions(
    draw: DrawFn,
    include_strings=True,
    include_floats=True,
    include_nones=True,
    include_bytes=True,
    include_datetimes=True,
    include_objectids=False,
    include_literals=True,
) -> tuple[Any, Any]:
    types = [bool, int]
    strats = {bool: booleans(), int: integers()}
    if include_strings:
        types.append(str)
        strats[str] = text()
    if include_floats:
        types.append(float)
        strats[float] = floats(allow_nan=False)
    if include_nones:
        types.append(None)
        strats[None] = just(None)
    if include_bytes:
        types.append(bytes)
        strats[bytes] = binary()
    if include_datetimes:
        types.append(datetime)
        strats[datetime] = datetimes(
            min_value=datetime(1970, 1, 1), max_value=datetime(2038, 1, 1)
        )
    if include_objectids:
        types.append(ObjectId)
        strats[ObjectId] = builds(ObjectId)

    chosen_types = draw(sets(sampled_from(types), min_size=2))

    if include_literals:
        from typing import Literal

        # We can replace some of the types with 1+ literal types.
        if str in chosen_types:
            strat = draw(sampled_from(["leave", "literal", "newtype"]))
            if strat == "literal":
                chosen_types.remove(str)
                vals = draw(sets(text(), min_size=1, max_size=2))
                for lit in vals:
                    t = Literal[lit]
                    chosen_types.add(t)
                    strats[t] = just(lit)
            elif strat == "newtype":
                chosen_types.remove(str)
                chosen_types.add(NewStr)
                strats[NewStr] = strats.pop(str)
        if bool in chosen_types:
            strat = draw(sampled_from(["leave", "literal", "newtype"]))
            if strat == "literal":
                chosen_types.remove(bool)
                val = draw(booleans())
                t = Literal[val]
                chosen_types.add(t)
                strats[t] = just(val)
            elif strat == "newtype":
                chosen_types.remove(bool)
                chosen_types.add(NewBool)
                strats[NewBool] = strats.pop(bool)
        if int in chosen_types:
            strat = draw(sampled_from(["leave", "literal", "newtype"]))
            if strat == "literal":
                chosen_types.remove(int)
                vals = draw(sets(integers(), min_size=1, max_size=2))
                for val in vals:
                    t = Literal[val]
                    chosen_types.add(t)
                    strats[t] = just(val)
            elif strat == "newtype":
                # NewTypes instead.
                chosen_types.remove(int)
                chosen_types.add(NewInt)
                strats[NewInt] = strats.pop(int)

    return Union[tuple(chosen_types)], draw(one_of(*[strats[t] for t in chosen_types]))


@given(everythings())
def test_stdlib_json(everything: Everything):
    converter = json_make_converter()
    assert (
        converter.structure(
            json_loads(json_dumps(converter.unstructure(everything))), Everything
        )
        == everything
    )


@given(everythings())
def test_stdlib_json_converter(everything: Everything):
    converter = json_make_converter()
    assert converter.loads(converter.dumps(everything), Everything) == everything


@given(everythings())
def test_stdlib_json_converter_unstruct_collection_overrides(everything: Everything):
    converter = json_make_converter(unstruct_collection_overrides={Set: sorted})
    raw = converter.unstructure(everything)
    assert raw["a_set"] == sorted(raw["a_set"])
    assert raw["a_mutable_set"] == sorted(raw["a_mutable_set"])
    assert raw["a_frozenset"] == sorted(raw["a_frozenset"])


@given(
    union_and_val=native_unions(include_bytes=False, include_datetimes=False),
    detailed_validation=...,
)
@settings(max_examples=1000)
def test_stdlib_json_unions(union_and_val: tuple, detailed_validation: bool):
    """Native union passthrough works."""
    converter = json_make_converter(detailed_validation=detailed_validation)
    type, val = union_and_val

    assert converter.structure(val, type) == val


@given(
    union_and_val=native_unions(include_strings=False, include_bytes=False),
    detailed_validation=...,
)
def test_stdlib_json_unions_with_spillover(
    union_and_val: tuple, detailed_validation: bool
):
    """Native union passthrough works and can handle spillover.

    The stdlib json converter cannot handle datetimes natively.
    """
    converter = json_make_converter(detailed_validation=detailed_validation)
    type, val = union_and_val

    assert converter.structure(converter.unstructure(val), type) == val


def test_stdlib_json_native_enums():
    """Bare, string and int enums are handled correctly."""
    converter = json_make_converter()
    assert (
        json_loads(converter.dumps(Everything.AnIntEnum.A))
        == Everything.AnIntEnum.A.value
    )
    assert (
        json_loads(converter.dumps(Everything.AStringEnum.A))
        == Everything.AStringEnum.A.value
    )
    assert (
        json_loads(converter.dumps(Everything.ABareEnum.B))
        == Everything.ABareEnum.B.value
    )


def test_stdlib_json_efficient_enum():
    """`str` and `int` enums are handled efficiently."""
    converter = json_make_converter()

    assert converter.get_unstructure_hook(Everything.AnIntEnum) == identity
    assert converter.get_unstructure_hook(Everything.AStringEnum) == identity
    assert converter.get_unstructure_hook(fields(Everything).a_literal) == identity


@given(
    everythings(
        min_int=-9223372036854775808, max_int=9223372036854775807, allow_inf=False
    )
)
def test_ujson(everything: Everything):
    from ujson import dumps as ujson_dumps
    from ujson import loads as ujson_loads

    converter = ujson_make_converter()
    raw = ujson_dumps(converter.unstructure(everything))
    assert converter.structure(ujson_loads(raw), Everything) == everything


@given(
    everythings(
        min_int=-9223372036854775808, max_int=9223372036854775807, allow_inf=False
    )
)
def test_ujson_converter(everything: Everything):
    converter = ujson_make_converter()
    raw = converter.dumps(everything)
    assert converter.loads(raw, Everything) == everything


@given(
    everythings(
        min_int=-9223372036854775808, max_int=9223372036854775807, allow_inf=False
    )
)
def test_ujson_converter_unstruct_collection_overrides(everything: Everything):
    converter = ujson_make_converter(unstruct_collection_overrides={Set: sorted})
    raw = converter.unstructure(everything)
    assert raw["a_set"] == sorted(raw["a_set"])
    assert raw["a_mutable_set"] == sorted(raw["a_mutable_set"])
    assert raw["a_frozenset"] == sorted(raw["a_frozenset"])


@given(
    union_and_val=native_unions(include_bytes=False, include_datetimes=False),
    detailed_validation=...,
)
def test_ujson_unions(union_and_val: tuple, detailed_validation: bool):
    """Native union passthrough works."""
    converter = ujson_make_converter(detailed_validation=detailed_validation)
    type, val = union_and_val

    assert converter.structure(val, type) == val


def test_ujson_native_enums():
    """Bare, string and int enums are handled correctly."""
    converter = ujson_make_converter()
    assert (
        json_loads(converter.dumps(Everything.AnIntEnum.A))
        == Everything.AnIntEnum.A.value
    )
    assert (
        json_loads(converter.dumps(Everything.AStringEnum.A))
        == Everything.AStringEnum.A.value
    )
    assert (
        json_loads(converter.dumps(Everything.ABareEnum.B))
        == Everything.ABareEnum.B.value
    )


def test_ujson_efficient_enum():
    """Bare, `str` and `int` enums are handled efficiently."""
    converter = ujson_make_converter()

    assert converter.get_unstructure_hook(Everything.AnIntEnum) == identity
    assert converter.get_unstructure_hook(Everything.AStringEnum) == identity
    assert converter.get_unstructure_hook(fields(Everything).a_literal.type) == identity


@pytest.mark.skipif(NO_ORJSON, reason="orjson not available")
@given(
    everythings(
        min_int=-9223372036854775808, max_int=9223372036854775807, allow_inf=False
    ),
    booleans(),
)
def test_orjson(everything: Everything, detailed_validation: bool):
    from orjson import dumps as orjson_dumps
    from orjson import loads as orjson_loads

    from cattrs.preconf.orjson import make_converter as orjson_make_converter

    converter = orjson_make_converter(detailed_validation=detailed_validation)
    raw = orjson_dumps(converter.unstructure(everything))
    assert converter.structure(orjson_loads(raw), Everything) == everything


@pytest.mark.skipif(NO_ORJSON, reason="orjson not available")
@given(
    everythings(
        min_int=-9223372036854775808, max_int=9223372036854775807, allow_inf=False
    ),
    booleans(),
)
def test_orjson_converter(everything: Everything, detailed_validation: bool):
    from cattrs.preconf.orjson import make_converter as orjson_make_converter

    converter = orjson_make_converter(detailed_validation=detailed_validation)
    raw = converter.dumps(everything)
    assert converter.loads(raw, Everything) == everything


@pytest.mark.skipif(NO_ORJSON, reason="orjson not available")
@given(
    everythings(
        min_int=-9223372036854775808, max_int=9223372036854775807, allow_inf=False
    )
)
def test_orjson_converter_unstruct_collection_overrides(everything: Everything):
    from cattrs.preconf.orjson import make_converter as orjson_make_converter

    converter = orjson_make_converter(unstruct_collection_overrides={Set: sorted})
    raw = converter.unstructure(everything)
    assert raw["a_set"] == sorted(raw["a_set"])
    assert raw["a_mutable_set"] == sorted(raw["a_mutable_set"])
    assert raw["a_frozenset"] == sorted(raw["a_frozenset"])


@pytest.mark.skipif(NO_ORJSON, reason="orjson not available")
@given(
    union_and_val=native_unions(include_bytes=False, include_datetimes=False),
    detailed_validation=...,
)
def test_orjson_unions(union_and_val: tuple, detailed_validation: bool):
    """Native union passthrough works."""
    from cattrs.preconf.orjson import make_converter as orjson_make_converter

    converter = orjson_make_converter(detailed_validation=detailed_validation)
    type, val = union_and_val

    assert converter.structure(val, type) == val


@pytest.mark.skipif(NO_ORJSON, reason="orjson not available")
def test_orjson_native_enums():
    """Bare, string and int enums are handled correctly."""
    from cattrs.preconf.orjson import make_converter as orjson_make_converter

    converter = orjson_make_converter()

    assert (
        json_loads(converter.dumps(Everything.AnIntEnum.A))
        == Everything.AnIntEnum.A.value
    )
    assert (
        json_loads(converter.dumps(Everything.AStringEnum.A))
        == Everything.AStringEnum.A.value
    )
    assert (
        json_loads(converter.dumps(Everything.ABareEnum.B))
        == Everything.ABareEnum.B.value
    )


@pytest.mark.skipif(NO_ORJSON, reason="orjson not available")
def test_orjson_efficient_enum():
    """Bare, `str` and `int` enums are handled efficiently."""
    from cattrs.preconf.orjson import make_converter as orjson_make_converter

    converter = orjson_make_converter()

    assert converter.get_unstructure_hook(Everything.AnIntEnum) == identity
    assert converter.get_unstructure_hook(Everything.AStringEnum) == identity
    assert converter.get_unstructure_hook(Everything.ABareEnum) == identity
    assert converter.get_unstructure_hook(fields(Everything).a_literal.type) == identity
    assert (
        converter.get_unstructure_hook(fields(Everything).a_literal_with_bare.type)
        == identity
    )


@given(everythings(min_int=-9223372036854775808, max_int=18446744073709551615))
def test_msgpack(everything: Everything):
    from msgpack import dumps as msgpack_dumps
    from msgpack import loads as msgpack_loads

    converter = msgpack_make_converter()
    raw = msgpack_dumps(converter.unstructure(everything))
    assert (
        converter.structure(msgpack_loads(raw, strict_map_key=False), Everything)
        == everything
    )


@given(everythings(min_int=-9223372036854775808, max_int=18446744073709551615))
def test_msgpack_converter(everything: Everything):
    converter = msgpack_make_converter()
    raw = converter.dumps(everything)
    assert converter.loads(raw, Everything, strict_map_key=False) == everything


@given(everythings(min_int=-9223372036854775808, max_int=18446744073709551615))
def test_msgpack_converter_unstruct_collection_overrides(everything: Everything):
    converter = msgpack_make_converter(unstruct_collection_overrides={Set: sorted})
    raw = converter.unstructure(everything)
    assert raw["a_set"] == sorted(raw["a_set"])
    assert raw["a_mutable_set"] == sorted(raw["a_mutable_set"])
    assert raw["a_frozenset"] == sorted(raw["a_frozenset"])


@given(union_and_val=native_unions(include_datetimes=False), detailed_validation=...)
def test_msgpack_unions(union_and_val: tuple, detailed_validation: bool):
    """Native union passthrough works."""
    converter = msgpack_make_converter(detailed_validation=detailed_validation)
    type, val = union_and_val

    assert converter.structure(val, type) == val


def test_msgpack_native_enums():
    """Bare, string and int enums are handled correctly."""

    converter = msgpack_make_converter()

    assert converter.dumps(Everything.AnIntEnum.A) == converter.dumps(
        Everything.AnIntEnum.A.value
    )
    assert converter.dumps(Everything.AStringEnum.A) == converter.dumps(
        Everything.AStringEnum.A.value
    )
    assert converter.dumps(Everything.ABareEnum.B) == converter.dumps(
        Everything.ABareEnum.B.value
    )


def test_msgpack_efficient_enum():
    """`str` and `int` enums are handled efficiently."""
    converter = msgpack_make_converter()

    assert converter.get_unstructure_hook(Everything.AnIntEnum) == identity
    assert converter.get_unstructure_hook(Everything.AStringEnum) == identity
    assert converter.get_unstructure_hook(fields(Everything).a_literal.type) == identity


@given(
    everythings(
        min_int=-9223372036854775808,
        max_int=9223372036854775807,
        allow_null_bytes_in_keys=False,
        allow_datetime_microseconds=False,
    ),
    booleans(),
)
def test_bson(everything: Everything, detailed_validation: bool):
    from bson import decode as bson_loads
    from bson import encode as bson_dumps

    converter = bson_make_converter(detailed_validation=detailed_validation)
    raw = bson_dumps(
        converter.unstructure(everything), codec_options=CodecOptions(tz_aware=True)
    )
    assert (
        converter.structure(
            bson_loads(raw, codec_options=CodecOptions(tz_aware=True)), Everything
        )
        == everything
    )


@given(
    everythings(
        min_int=-9223372036854775808,
        max_int=9223372036854775807,
        allow_null_bytes_in_keys=False,
        allow_datetime_microseconds=False,
    ),
    booleans(),
)
def test_bson_converter(everything: Everything, detailed_validation: bool):
    converter = bson_make_converter(detailed_validation=detailed_validation)
    raw = converter.dumps(everything, codec_options=CodecOptions(tz_aware=True))
    assert (
        converter.loads(raw, Everything, codec_options=CodecOptions(tz_aware=True))
        == everything
    )


@given(
    everythings(
        min_int=-9223372036854775808,
        max_int=9223372036854775807,
        allow_null_bytes_in_keys=False,
        allow_datetime_microseconds=False,
    )
)
def test_bson_converter_unstruct_collection_overrides(everything: Everything):
    converter = bson_make_converter(unstruct_collection_overrides={Set: sorted})
    raw = converter.unstructure(everything)
    assert raw["a_set"] == sorted(raw["a_set"])
    assert raw["a_mutable_set"] == sorted(raw["a_mutable_set"])
    assert raw["a_frozenset"] == sorted(raw["a_frozenset"])


@given(union_and_val=native_unions(include_objectids=True), detailed_validation=...)
def test_bson_unions(union_and_val: tuple, detailed_validation: bool):
    """Native union passthrough works."""
    converter = bson_make_converter(detailed_validation=detailed_validation)
    type, val = union_and_val

    assert converter.structure(val, type) == val


def test_bson_objectid():
    """BSON ObjectIds are supported by default."""
    converter = bson_make_converter()
    o = ObjectId()
    assert o == converter.structure(str(o), ObjectId)
    assert o == converter.structure(o, ObjectId)


def test_bson_native_enums():
    """Bare, string and int enums are handled correctly."""

    converter = bson_make_converter()

    assert converter.dumps({"a": Everything.AnIntEnum.A}) == converter.dumps(
        {"a": Everything.AnIntEnum.A.value}
    )
    assert converter.dumps({"a": Everything.AStringEnum.A}) == converter.dumps(
        {"a": Everything.AStringEnum.A.value}
    )
    assert converter.dumps({"a": Everything.ABareEnum.B}) == converter.dumps(
        {"a": Everything.ABareEnum.B.value}
    )


def test_bson_efficient_enum():
    """`str` and `int` enums are handled efficiently."""
    converter = bson_make_converter()

    assert converter.get_unstructure_hook(Everything.AnIntEnum) == identity
    assert converter.get_unstructure_hook(Everything.AStringEnum) == identity
    assert converter.get_unstructure_hook(fields(Everything).a_literal.type) == identity


@given(
    everythings(
        min_key_length=1,
        allow_null_bytes_in_keys=False,
        key_blacklist_characters=['"', "\\"],
        allow_control_characters_in_values=False,
    ),
    booleans(),
)
def test_tomlkit(everything: Everything, detailed_validation: bool):
    from tomlkit import dumps as tomlkit_dumps
    from tomlkit import loads as tomlkit_loads

    converter = tomlkit_make_converter(detailed_validation=detailed_validation)
    unstructured = converter.unstructure(everything)
    raw = tomlkit_dumps(unstructured)

    loaded = tomlkit_loads(raw)
    # Check if we're unstructuring dates to native toml dates
    assert isinstance(loaded["a_date"], date)

    assert converter.structure(loaded, Everything) == everything


@given(
    everythings(
        min_key_length=1,
        allow_null_bytes_in_keys=False,
        key_blacklist_characters=['"', "\\"],
        allow_control_characters_in_values=False,
    ),
    booleans(),
)
def test_tomlkit_converter(everything: Everything, detailed_validation: bool):
    converter = tomlkit_make_converter(detailed_validation=detailed_validation)
    raw = converter.dumps(everything)
    assert converter.loads(raw, Everything) == everything


@given(
    everythings(
        min_key_length=1,
        allow_null_bytes_in_keys=False,
        key_blacklist_characters=['"', "\\"],
        allow_control_characters_in_values=False,
    )
)
def test_tomlkit_converter_unstruct_collection_overrides(everything: Everything):
    converter = tomlkit_make_converter(unstruct_collection_overrides={Set: sorted})
    raw = converter.unstructure(everything)
    assert raw["a_set"] == sorted(raw["a_set"])
    assert raw["a_mutable_set"] == sorted(raw["a_mutable_set"])
    assert raw["a_frozenset"] == sorted(raw["a_frozenset"])


@given(
    union_and_val=native_unions(
        include_nones=False, include_bytes=False, include_datetimes=False
    ),
    detailed_validation=...,
)
def test_tomlkit_unions(union_and_val: tuple, detailed_validation: bool):
    """Native union passthrough works."""
    converter = tomlkit_make_converter(detailed_validation=detailed_validation)
    type, val = union_and_val

    assert converter.structure(val, type) == val


def test_tomlkit_date_strings():
    """Dates represented as strings in toml work."""
    converter = tomlkit_make_converter()

    @define
    class A:
        a_date: date

    data = 'a_date = "2023-01-01"'
    assert converter.loads(data, A) == A(date(2023, 1, 1))


@given(
    everythings(
        allow_null_bytes_in_keys=False,
        key_blacklist_characters=['"', "\\"],
        allow_control_characters_in_values=False,
    ),
    booleans(),
)
def test_tomllib_converter(everything: Everything, detailed_validation: bool):

    converter = tomllib_make_converter(detailed_validation=detailed_validation)
    raw = converter.dumps(everything)

    assert converter.loads(raw, Everything) == everything


@given(
    everythings(
        allow_null_bytes_in_keys=False,
        key_blacklist_characters=['"', "\\"],
        allow_control_characters_in_values=False,
    ),
    booleans(),
)
def test_tomllib_converter_dumps(everything: Everything, detailed_validation: bool):
    converter = tomllib_make_converter(detailed_validation=detailed_validation)
    raw = converter.dumps(everything)
    assert converter.loads(raw, Everything) == everything


@given(
    everythings(
        allow_null_bytes_in_keys=False,
        key_blacklist_characters=['"', "\\"],
        allow_control_characters_in_values=False,
    )
)
def test_tomllib_converter_unstruct_collection_overrides(everything: Everything):
    converter = tomllib_make_converter(unstruct_collection_overrides={Set: sorted})
    raw = converter.unstructure(everything)
    assert raw["a_set"] == sorted(raw["a_set"])
    assert raw["a_mutable_set"] == sorted(raw["a_mutable_set"])
    assert raw["a_frozenset"] == sorted(raw["a_frozenset"])


@given(everythings(min_int=-9223372036854775808, max_int=18446744073709551615))
def test_cbor2(everything: Everything):
    from cbor2 import dumps as cbor2_dumps
    from cbor2 import loads as cbor2_loads

    converter = cbor2_make_converter()
    raw = cbor2_dumps(converter.unstructure(everything))
    assert converter.structure(cbor2_loads(raw), Everything) == everything


@given(everythings(min_int=-9223372036854775808, max_int=18446744073709551615))
def test_cbor2_converter(everything: Everything):
    converter = cbor2_make_converter()
    raw = converter.dumps(everything)
    assert converter.loads(raw, Everything) == everything


@given(everythings(min_int=-9223372036854775808, max_int=18446744073709551615))
def test_cbor2_converter_unstruct_collection_overrides(everything: Everything):
    converter = cbor2_make_converter(unstruct_collection_overrides={Set: sorted})
    raw = converter.unstructure(everything)
    assert raw["a_set"] == sorted(raw["a_set"])
    assert raw["a_mutable_set"] == sorted(raw["a_mutable_set"])
    assert raw["a_frozenset"] == sorted(raw["a_frozenset"])


@given(union_and_val=native_unions(include_datetimes=False), detailed_validation=...)
def test_cbor2_unions(union_and_val: tuple, detailed_validation: bool):
    """Native union passthrough works."""
    converter = cbor2_make_converter(detailed_validation=detailed_validation)
    type, val = union_and_val

    assert converter.structure(val, type) == val


def test_cbor2_native_enums():
    """Bare, string and int enums are handled correctly."""

    converter = cbor2_make_converter()

    assert converter.dumps(Everything.AnIntEnum.A) == converter.dumps(
        Everything.AnIntEnum.A.value
    )
    assert converter.dumps(Everything.AStringEnum.A) == converter.dumps(
        Everything.AStringEnum.A.value
    )
    assert converter.dumps(Everything.ABareEnum.B) == converter.dumps(
        Everything.ABareEnum.B.value
    )


def test_cbor2_efficient_enum():
    """`str` and `int` enums are handled efficiently."""
    converter = cbor2_make_converter()

    assert converter.get_unstructure_hook(Everything.AnIntEnum) == identity
    assert converter.get_unstructure_hook(Everything.AStringEnum) == identity
    assert converter.get_unstructure_hook(fields(Everything).a_literal.type) == identity


@pytest.mark.skipif(NO_MSGSPEC, reason="msgspec not available")
@given(everythings(allow_inf=False))
def test_msgspec_json_converter(everything: Everything):
    from cattrs.preconf.msgspec import make_converter as msgspec_make_converter

    converter = msgspec_make_converter()
    raw = converter.dumps(everything)
    assert converter.loads(raw, Everything) == everything


@pytest.mark.skipif(NO_MSGSPEC, reason="msgspec not available")
@given(everythings(allow_inf=False))
def test_msgspec_json_unstruct_collection_overrides(everything: Everything):
    """Ensure collection overrides work."""
    from cattrs.preconf.msgspec import make_converter as msgspec_make_converter

    converter = msgspec_make_converter(unstruct_collection_overrides={Set: sorted})
    raw = converter.unstructure(everything)
    assert raw["a_set"] == sorted(raw["a_set"])
    assert raw["a_mutable_set"] == sorted(raw["a_mutable_set"])
    assert raw["a_frozenset"] == sorted(raw["a_frozenset"])


@pytest.mark.skipif(NO_MSGSPEC, reason="msgspec not available")
@given(
    union_and_val=native_unions(include_datetimes=False, include_bytes=False),
    detailed_validation=...,
)
def test_msgspec_json_unions(union_and_val: tuple, detailed_validation: bool):
    """Native union passthrough works."""
    from cattrs.preconf.msgspec import make_converter as msgspec_make_converter

    converter = msgspec_make_converter(detailed_validation=detailed_validation)
    type, val = union_and_val

    assert converter.structure(val, type) == val


@pytest.mark.skipif(NO_MSGSPEC, reason="msgspec not available")
def test_msgspec_native_enums():
    """Bare, string and int enums are handled correctly."""
    from cattrs.preconf.msgspec import make_converter as msgspec_make_converter

    converter = msgspec_make_converter()

    assert converter.dumps(Everything.AnIntEnum.A) == converter.dumps(
        Everything.AnIntEnum.A.value
    )
    assert converter.dumps(Everything.AStringEnum.A) == converter.dumps(
        Everything.AStringEnum.A.value
    )
    assert converter.dumps(Everything.ABareEnum.B) == converter.dumps(
        Everything.ABareEnum.B.value
    )


@pytest.mark.skipif(NO_MSGSPEC, reason="msgspec not available")
def test_msgspec_efficient_enum():
    """Bare, `str` and `int` enums are handled efficiently."""
    from cattrs.preconf.msgspec import make_converter as msgspec_make_converter

    converter = msgspec_make_converter()

    assert converter.get_unstructure_hook(Everything.AnIntEnum) == identity
    assert converter.get_unstructure_hook(Everything.AStringEnum) == identity
    assert converter.get_unstructure_hook(Everything.ABareEnum) == identity
    assert converter.get_unstructure_hook(fields(Everything).a_literal.type) == identity
    assert (
        converter.get_unstructure_hook(fields(Everything).a_literal_with_bare.type)
        == identity
    )


@pytest.mark.parametrize(
    "converter_factory",
    [
        bson_make_converter,
        cbor2_make_converter,
        json_make_converter,
        msgpack_make_converter,
        tomlkit_make_converter,
        ujson_make_converter,
        pyyaml_make_converter,
    ],
)
def test_literal_dicts(converter_factory: Callable[[], Converter]):
    """Dicts with keys that aren't subclasses of `type` work."""
    converter = converter_factory()

    assert converter.structure({"a": 1}, Dict[Literal["a"], int]) == {"a": 1}
    assert converter.unstructure({"a": 1}, Dict[Literal["a"], int]) == {"a": 1}


@pytest.mark.skipif(NO_ORJSON, reason="orjson not available")
def test_literal_dicts_orjson():
    """Dicts with keys that aren't subclasses of `type` work."""
    from cattrs.preconf.orjson import make_converter as orjson_make_converter

    test_literal_dicts(orjson_make_converter)


@pytest.mark.skipif(NO_MSGSPEC, reason="msgspec not available")
def test_literal_dicts_msgspec():
    """Dicts with keys that aren't subclasses of `type` work."""
    from cattrs.preconf.msgspec import make_converter as msgspec_make_converter

    test_literal_dicts(msgspec_make_converter)


def test_literal_dicts_tomllib():
    """Dicts with keys that aren't subclasses of `type` work."""
    test_literal_dicts(tomllib_make_converter)



"""Tests for `typing.Self`."""

from dataclasses import dataclass
from typing import NamedTuple, Optional, TypedDict

from attrs import define
from typing_extensions import Self

from cattrs import Converter
from cattrs.cols import (
    namedtuple_dict_structure_factory,
    namedtuple_dict_unstructure_factory,
)


@define
class WithSelf:
    myself: Optional[Self]
    myself_with_default: Optional[Self] = None


@define
class WithSelfSubclass(WithSelf):
    pass


@dataclass
class WithSelfDataclass:
    myself: Optional[Self]


@dataclass
class WithSelfDataclassSubclass(WithSelfDataclass):
    pass


@define
class WithListOfSelf:
    myself: Optional[Self]
    selves: list[WithSelf]


class WithSelfTypedDict(TypedDict):
    field: int
    myself: Optional[Self]


class WithSelfNamedTuple(NamedTuple):
    myself: Optional[Self]


def test_self_roundtrip(genconverter):
    """A simple roundtrip works."""
    initial = WithSelf(WithSelf(None, WithSelf(None)))
    raw = genconverter.unstructure(initial)

    assert raw == {
        "myself": {
            "myself": None,
            "myself_with_default": {"myself": None, "myself_with_default": None},
        },
        "myself_with_default": None,
    }

    assert genconverter.structure(raw, WithSelf) == initial


def test_self_roundtrip_dataclass(genconverter):
    """A simple roundtrip works for dataclasses."""
    initial = WithSelfDataclass(WithSelfDataclass(None))
    raw = genconverter.unstructure(initial)

    assert raw == {"myself": {"myself": None}}

    assert genconverter.structure(raw, WithSelfDataclass) == initial


def test_self_roundtrip_typeddict(genconverter):
    """A simple roundtrip works for TypedDicts."""
    genconverter.register_unstructure_hook(int, str)

    initial: WithSelfTypedDict = {"field": 1, "myself": {"field": 2, "myself": None}}
    raw = genconverter.unstructure(initial)

    assert raw == {"field": "1", "myself": {"field": "2", "myself": None}}

    assert genconverter.structure(raw, WithSelfTypedDict) == initial


def test_self_roundtrip_namedtuple(genconverter):
    """A simple roundtrip works for NamedTuples."""
    genconverter.register_unstructure_hook_factory(
        lambda t: t is WithSelfNamedTuple, namedtuple_dict_unstructure_factory
    )
    genconverter.register_structure_hook_factory(
        lambda t: t is WithSelfNamedTuple, namedtuple_dict_structure_factory
    )

    initial = WithSelfNamedTuple(WithSelfNamedTuple(None))
    raw = genconverter.unstructure(initial)

    assert raw == {"myself": {"myself": None}}

    assert genconverter.structure(raw, WithSelfNamedTuple) == initial


def test_subclass_roundtrip(genconverter):
    """A simple roundtrip works for a dataclass subclass."""
    initial = WithSelfSubclass(WithSelfSubclass(None))
    raw = genconverter.unstructure(initial)

    assert raw == {
        "myself": {"myself": None, "myself_with_default": None},
        "myself_with_default": None,
    }

    assert genconverter.structure(raw, WithSelfSubclass) == initial


def test_subclass_roundtrip_dataclass(genconverter):
    """A simple roundtrip works for a dataclass subclass."""
    initial = WithSelfDataclassSubclass(WithSelfDataclassSubclass(None))
    raw = genconverter.unstructure(initial)

    assert raw == {"myself": {"myself": None}}

    assert genconverter.structure(raw, WithSelfDataclassSubclass) == initial


def test_nested_roundtrip(genconverter: Converter):
    """A more complex roundtrip, with several Self classes."""
    initial = WithListOfSelf(WithListOfSelf(None, []), [WithSelf(WithSelf(None))])
    raw = genconverter.unstructure(initial)

    assert raw == {
        "myself": {"myself": None, "selves": []},
        "selves": [
            {
                "myself": {"myself": None, "myself_with_default": None},
                "myself_with_default": None,
            }
        ],
    }

    assert genconverter.structure(raw, WithListOfSelf) == initial



"""Tests for the `cattrs.cols` module."""

from collections.abc import MutableSequence, Sequence, Set
from typing import Dict

from immutables import Map

from cattrs import BaseConverter, Converter
from cattrs._compat import FrozenSet
from cattrs.cols import (
    is_abstract_set,
    is_any_set,
    is_sequence,
    iterable_unstructure_factory,
    list_structure_factory,
    mapping_unstructure_factory,
)

from ._compat import is_py310_plus


def test_set_overriding(converter: BaseConverter):
    """Overriding abstract sets by wrapping the default factory works."""

    converter.register_unstructure_hook_factory(
        is_any_set,
        lambda t, c: iterable_unstructure_factory(t, c, unstructure_to=sorted),
    )

    assert converter.unstructure({"c", "b", "a"}, Set[str]) == ["a", "b", "c"]
    assert converter.unstructure(frozenset(["c", "b", "a"]), FrozenSet[str]) == [
        "a",
        "b",
        "c",
    ]


def test_structuring_immutables_map(genconverter: Converter):
    """This should work due to our new is_mapping predicate."""
    assert genconverter.structure({"a": 1}, Map[str, int]) == Map(a=1)


def test_mapping_unstructure_direct(genconverter: Converter):
    """Some cases reduce to just `dict`."""
    assert genconverter.get_unstructure_hook(Dict[str, int]) is dict

    # `dict` is equivalent to `dict[Any, Any]`, which should not reduce to
    # just `dict`.
    assert genconverter.get_unstructure_hook(dict) is not dict

    if is_py310_plus:
        assert genconverter.get_unstructure_hook(dict[str, int]) is dict


def test_mapping_unstructure_to(genconverter: Converter):
    """`unstructure_to` works."""
    hook = mapping_unstructure_factory(Dict[str, str], genconverter, unstructure_to=Map)
    assert hook({"a": "a"}).__class__ is Map


def test_structure_sequences(converter: BaseConverter):
    """Sequences are structured to tuples."""

    assert converter.structure(["1", 2, 3.0], Sequence[int]) == (1, 2, 3)


def test_structure_sequences_override(converter: BaseConverter):
    """Sequences can be overriden to structure to lists, as previously."""

    converter.register_structure_hook_factory(is_sequence, list_structure_factory)

    assert converter.structure(["1", 2, 3.0], Sequence[int]) == [1, 2, 3]


def test_structure_mut_sequences(converter: BaseConverter):
    """Mutable sequences are structured to lists."""

    assert converter.structure(["1", 2, 3.0], MutableSequence[int]) == [1, 2, 3]


def test_abstract_set_predicate():
    """`is_abstract_set` works."""

    assert is_abstract_set(Set)
    assert is_abstract_set(Set[str])

    assert not is_abstract_set(set)
    assert not is_abstract_set(set[str])


def test_structure_abstract_sets(converter: BaseConverter):
    """Abstract sets structure to frozensets."""

    assert converter.structure(["1", "2", "3"], Set[int]) == frozenset([1, 2, 3])
    assert isinstance(converter.structure([1, 2, 3], Set[int]), frozenset)


def test_structure_abstract_sets_override(converter: BaseConverter):
    """Abstract sets can be overridden to structure to mutable sets, as before."""
    converter.register_structure_hook_func(is_abstract_set, converter._structure_set)

    assert converter.structure(["1", 2, 3.0], Set[int]) == {1, 2, 3}



"""Helpers for tests."""

from typing import Any


def assert_only_unstructured(obj: Any):
    """Assert the object is comprised of only unstructured data:

    * dicts, lists, tuples
    * strings, ints, floats, bools, None
    """
    if isinstance(obj, dict):
        for k, v in obj.items():
            assert_only_unstructured(k)
            assert_only_unstructured(v)
    elif isinstance(obj, (list, tuple, frozenset, set)):
        for e in obj:
            assert_only_unstructured(e)
    else:
        assert isinstance(obj, (int, float, str, bool, type(None)))



from enum import Enum, IntEnum, StrEnum

import pytest

from cattrs import BaseConverter, Converter
from cattrs.preconf.msgspec import MsgspecJsonConverter
from cattrs.preconf.orjson import OrjsonConverter


class SimpleEnum(Enum):
    FIRST = "first"
    SECOND = "second"
    THIRD = "third"


class SimpleIntEnum(IntEnum):
    ONE = 1
    TWO = 2
    THREE = 3


class SimpleStrEnum(StrEnum):
    ALPHA = "alpha"
    BETA = "beta"
    GAMMA = "gamma"


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter, MsgspecJsonConverter, OrjsonConverter])
def test_unstructure_simple_enum(benchmark, converter_cls):
    """Benchmark unstructuring a simple enum."""
    c = converter_cls()
    benchmark(c.unstructure, SimpleEnum.FIRST)


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter, MsgspecJsonConverter, OrjsonConverter])
def test_structure_simple_enum(benchmark, converter_cls):
    """Benchmark structuring a simple enum."""
    c = converter_cls()
    benchmark(c.structure, "first", SimpleEnum)


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter, MsgspecJsonConverter, OrjsonConverter])
def test_unstructure_simple_int_enum(benchmark, converter_cls):
    """Benchmark unstructuring a simple IntEnum."""
    c = converter_cls()
    benchmark(c.unstructure, SimpleIntEnum.ONE)


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter, MsgspecJsonConverter, OrjsonConverter])
def test_structure_simple_int_enum(benchmark, converter_cls):
    """Benchmark structuring a simple IntEnum."""
    c = converter_cls()
    benchmark(c.structure, 1, SimpleIntEnum)


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter, MsgspecJsonConverter, OrjsonConverter])
def test_unstructure_simple_str_enum(benchmark, converter_cls):
    """Benchmark unstructuring a simple StrEnum."""
    c = converter_cls()
    benchmark(c.unstructure, SimpleStrEnum.ALPHA)


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter, MsgspecJsonConverter, OrjsonConverter])
def test_structure_simple_str_enum(benchmark, converter_cls):
    """Benchmark structuring a simple StrEnum."""
    c = converter_cls()
    benchmark(c.structure, "alpha", SimpleStrEnum)

